import torch
import torch.nn as nn
import numpy as np

from .masked_losses import masked_l1_loss


def compute_scale_and_shift(prediction, target, mask):
    # system matrix: A = [[a_00, a_01], [a_10, a_11]]
    a_00 = torch.sum(mask * prediction * prediction, (1, 2))
    a_01 = torch.sum(mask * prediction, (1, 2))
    a_11 = torch.sum(mask, (1, 2))

    # right hand side: b = [b_0, b_1]
    b_0 = torch.sum(mask * prediction * target, (1, 2))
    b_1 = torch.sum(mask * target, (1, 2))

    # solution: x = A^-1 . b = [[a_11, -a_01], [-a_10, a_00]] / (a_00 * a_11 - a_01 * a_10) . b
    x_0 = torch.zeros_like(b_0)
    x_1 = torch.zeros_like(b_1)

    det = a_00 * a_11 - a_01 * a_01
    valid = det.nonzero()

    x_0[valid] = (a_11[valid] * b_0[valid] - a_01[valid] * b_1[valid]) / (det[valid] + 1e-6)
    x_1[valid] = (-a_01[valid] * b_0[valid] + a_00[valid] * b_1[valid]) / (det[valid] + 1e-6)

    return x_0, x_1


def masked_shift_and_scale(depth_preds, depth_gt, mask_valid):
    depth_preds_nan = depth_preds.clone()
    depth_gt_nan = depth_gt.clone()
    depth_preds_nan[~mask_valid] = np.nan
    depth_gt_nan[~mask_valid] = np.nan

    mask_diff = mask_valid.view(mask_valid.size()[:2] + (-1,)).sum(-1, keepdims=True) + 1

    t_gt = depth_gt_nan.view(depth_gt_nan.size()[:2] + (-1,)).nanmedian(-1, keepdims=True)[0].unsqueeze(-1)
    t_gt[torch.isnan(t_gt)] = 0
    diff_gt = torch.abs(depth_gt - t_gt)
    diff_gt[~mask_valid] = 0
    s_gt = (diff_gt.view(diff_gt.size()[:2] + (-1,)).sum(-1, keepdims=True) / mask_diff).unsqueeze(-1)
    depth_gt_aligned = (depth_gt - t_gt) / (s_gt + 1e-6)


    t_pred = depth_preds_nan.view(depth_preds_nan.size()[:2] + (-1,)).nanmedian(-1, keepdims=True)[0].unsqueeze(-1)
    t_pred[torch.isnan(t_pred)] = 0
    diff_pred = torch.abs(depth_preds - t_pred)
    diff_pred[~mask_valid] = 0
    s_pred = (diff_pred.view(diff_pred.size()[:2] + (-1,)).sum(-1, keepdims=True) / mask_diff).unsqueeze(-1)
    depth_pred_aligned = (depth_preds - t_pred) / (s_pred + 1e-6)

    return depth_pred_aligned, depth_gt_aligned


def reduction_batch_based(image_loss, M):
    # average of all valid pixels of the batch

    # avoid division by 0 (if sum(M) = sum(sum(mask)) = 0: sum(image_loss) = 0)
    divisor = torch.sum(M)

    if divisor == 0:
        return 0
    else:
        return torch.sum(image_loss) / divisor


def reduction_image_based(image_loss, M):
    # mean of average of valid pixels of an image

    # avoid division by 0 (if M = sum(mask) = 0: image_loss = 0)
    valid = M.nonzero()

    image_loss[valid] = image_loss[valid] / M[valid]

    return torch.mean(image_loss)



def gradient_loss(prediction, target, mask, reduction=reduction_batch_based):

    M = torch.sum(mask, (1, 2))

    diff = prediction - target
    diff = torch.mul(mask, diff)

    grad_x = torch.abs(diff[:, :, 1:] - diff[:, :, :-1])
    mask_x = torch.mul(mask[:, :, 1:], mask[:, :, :-1])
    grad_x = torch.mul(mask_x, grad_x)

    grad_y = torch.abs(diff[:, 1:, :] - diff[:, :-1, :])
    mask_y = torch.mul(mask[:, 1:, :], mask[:, :-1, :])
    grad_y = torch.mul(mask_y, grad_y)

    image_loss = torch.sum(grad_x, (1, 2)) + torch.sum(grad_y, (1, 2))

    return reduction(image_loss, M)



class SSIMAE(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, depth_preds, depth_gt, mask_valid):
        depth_pred_aligned, depth_gt_aligned = masked_shift_and_scale(depth_preds, depth_gt, mask_valid)
        ssi_mae_loss = masked_l1_loss(depth_pred_aligned, depth_gt_aligned, mask_valid)
        return ssi_mae_loss


class GradientMatchingTerm(nn.Module):
    def __init__(self, scales=4, reduction='batch-based'):
        super().__init__()

        if reduction == 'batch-based':
            self.__reduction = reduction_batch_based
        else:
            self.__reduction = reduction_image_based

        self.__scales = scales

    def forward(self, prediction, target, mask):
        total = 0

        for scale in range(self.__scales):
            step = pow(2, scale)

            total += gradient_loss(prediction[:, ::step, ::step], target[:, ::step, ::step],
                                   mask[:, ::step, ::step], reduction=self.__reduction)

        return total


class MidasLoss(nn.Module):
    def __init__(self, alpha=0.1, scales=4, reduction='image-based'):
        super().__init__()

        self.__ssi_mae_loss = SSIMAE()
        self.__gradient_matching_term = GradientMatchingTerm(scales=scales, reduction=reduction)
        self.__alpha = alpha
        self.__prediction_ssi = None

    def forward(self, prediction, target, mask):
        prediction_inverse = 1 / (prediction.squeeze(1)+1e-6)
        target_inverse = 1 / (target.squeeze(1)+1e-6)
        ssi_loss = self.__ssi_mae_loss(prediction, target, mask)

        scale, shift = compute_scale_and_shift(prediction_inverse, target_inverse, mask.squeeze(1))
        self.__prediction_ssi = scale.view(-1, 1, 1) * prediction_inverse + shift.view(-1, 1, 1)
        reg_loss = self.__gradient_matching_term(self.__prediction_ssi, target_inverse, mask.squeeze(1))
        if self.__alpha > 0:
            total = ssi_loss + self.__alpha * reg_loss

        return total, ssi_loss, reg_loss



import torch
import numpy as np

def masked_l1_loss(preds, target, mask_valid):
    element_wise_loss = abs(preds - target)
    element_wise_loss[~mask_valid] = 0
    return element_wise_loss.sum() / mask_valid.sum() 

def masked_mse_loss(preds, target, mask_valid):
    element_wise_loss = (preds - target)**2
    element_wise_loss[~mask_valid] = 0
    return element_wise_loss.sum() / mask_valid.sum()

def masked_cosine_angular_loss(preds, target, mask_valid):
    preds = (2 * preds - 1).clamp(-1, 1)
    target = (2 * target - 1).clamp(-1, 1)
    mask_valid = mask_valid[:,0,:,:].bool().squeeze(1)
    preds = preds.permute(0,2,3,1)[mask_valid, :]
    target = target.permute(0,2,3,1)[mask_valid, :]
    preds_norm =  torch.nn.functional.normalize(preds, p=2, dim=1)
    target_norm = torch.nn.functional.normalize(target, p=2, dim=1)
    loss = torch.mean(-torch.sum(preds_norm * target_norm, dim = 1))
    return loss


def masked_loss(element_wise_loss, mask_valid):
    element_wise_loss[~mask_valid] = 0
    if mask_valid.sum() == 0:
        return torch.tensor(0.0).to(element_wise_loss.device)
    return element_wise_loss.sum() / mask_valid.sum()


import random
import torch.nn.functional as F
import sys
import os
sys.path.append(os.path.dirname(os.path.abspath(__file__)))
sys.path.append('../')
from refocus_augmentation import RefocusImageAugmentation
try:
    from kornia.augmentation import *
except:
    print("Error importing kornia augmentation")


class Augmentation:
    def __init__(self):
        pass
        # self.refocus_aug = RefocusImageAugmentation(10, 0.005, 6.0, return_segments=False)

    def augment_rgb(self, batch):
            rgb = batch['positive']['rgb']

            # color jitter
            jitter = ColorJitter(
                (0.1, 0.5), 
                (0.1, 0.5), 
                (0.1, 0.5), 
                (-0.05, 0.05), 
                p=.7)
            
            # augmented_rgb = jitter(rgb)
            augmented_rgb = rgb
            
            # blur
            p = random.random()
            if p < 0.7:
                p = random.random()
                if p < 0.5:
                    aug = RandomSharpness(.3, p=1.)
                    augmented_rgb = aug(augmented_rgb)

                p = random.random()
                if p < 0.5:
                    aug = RandomMotionBlur((3, 7), random.uniform(10., 50.), 0.5, p=1.)
                    augmented_rgb = aug(augmented_rgb)

                p = random.random()
                if p < 0.1:
                    aug = RandomGaussianBlur((7, 7), (0.1, 2.0), p=1.)
                    augmented_rgb = aug(augmented_rgb)
                elif p < 0.4:
                    aug = RandomGaussianBlur((5, 5), (0.1, 2.0), p=1.)
                    augmented_rgb = aug(augmented_rgb)
                elif p < 0.6:
                    aug = RandomGaussianBlur((3, 3), (0.1, 2.0), p=1.)
                    augmented_rgb = aug(augmented_rgb)

                # elif p < 0.8:
                #     depth = batch['positive']['depth_euclidean']
                #     if depth[depth < 1.0].shape[0] != 0:
                #         depth[depth >= 1.0] = depth[depth < 1.0].max()
                #     else:
                #         depth[depth >= 1.0] = 0.99
                #         print("**")
                #     augmented_rgb = self.refocus_aug(augmented_rgb, depth)
                

            return augmented_rgb

    def resize_augmentation(self, batch, tasks, fixed_size=None):
        p = random.random()

        if p < 0.4:
            resize_method = 'centercrop'
        elif p < 0.7:
            resize_method = 'randomcrop'
        else:
            resize_method = 'resize'

        if fixed_size is not None: 
            h = fixed_size
            w = fixed_size
        else:
            img_sizes = [256, 320, 384, 448, 512]
            while True:
                h = random.choice(img_sizes)
                w = random.choice(img_sizes)
                if resize_method == 'resize':
                    if h < 1.5 * w and w < 1.5 * h: break
                else:   
                    if h < 2 * w and w < 2 * h: break


        if resize_method == 'randomcrop':
            min_x, min_y = 0, 0
            size_x, size_y = batch[tasks[0]].shape[-2], batch[tasks[0]].shape[-1]
            if size_x != h:
                min_x = random.randrange(0, size_x - h - 2)
            if size_y != w:
                min_y = random.randrange(0, size_y - w - 2)

        for task in tasks:
            if len(batch[task].shape) == 3:
                batch[task] = batch[task].unsqueeze(axis=0)

            if resize_method == 'centercrop':
                centercrop = CenterCrop((h, w), p=1.)
                batch[task] = centercrop(batch[task])

            elif resize_method == 'randomcrop':
                batch[task] = batch[task][:, :, min_x:min_x + h, min_y:min_y + w]

            elif resize_method == 'resize':

                if task == 'rgb':
                    batch[task] = F.interpolate(batch[task], (h, w), mode='bilinear')
                else:
                    batch[task] = F.interpolate(batch[task], (h, w), mode='nearest')

        return batch


####################
# Tasks
####################

task_parameters = {
    'autoencoding':{
        'out_channels': 3,

    },

    'denoising':{
        'out_channels': 3,

    },
    'colorization': {
        'out_channels': 3,

    },
    'class_object':{
        'out_channels': 1000,

    },
    'class_scene':{
        'out_channels': 365,

    },
    'depth_zbuffer':{
        'out_channels': 1,
        'mask_val': 1.0,
        'clamp_to': (0.0, 8000.0 / (2**16 - 1)) # Same as consistency
    },
    'depth_euclidean':{
        'out_channels': 1,
        'clamp_to': (0.0, 8000.0 / (2**16 - 1)) # Same as consistency
#         'mask_val': 1.0,
    },
    'edge_texture': {
        'out_channels': 1,
        'clamp_to': (0.0, 0.25)
    },
    'edge_occlusion': {
        'out_channels': 1,

    },
    'inpainting':{
        'out_channels': 3,

    },
    'keypoints3d': {
        'out_channels': 1,

    },
    'keypoints2d':{
        'out_channels': 1,

    },
    'principal_curvature':{
        'out_channels': 2,
        'mask_val': 0.0,
    },

    'reshading':{
        'out_channels': 1,

    }, 
    'normal':{
        'out_channels': 3,
#         'mask_val': 0.004,
        'mask_val': 0.502,
    },
    'mask_valid':{
        'out_channels': 1,
        'mask_val': 0.0,
    },
    'rgb':{
        'out_channels': 3,
    },
    'segment_semantic': {
        'out_channels': 17,
    },
    'segment_unsup2d': {
        'out_channels': 64,
    },
    'segment_unsup25d': {
        'out_channels': 64,
    },
    'segment_instance': {
    },
    'segment_panoptic': {
        'out_channels': 2,
    },
    'fragments': {
        'out_channels': 1
    }
}

        
PIX_TO_PIX_TASKS = ['colorization', 'edge_texture', 'edge_occlusion',  'keypoints3d', 'keypoints2d', 'reshading', 'depth_zbuffer', 'depth_euclidean', 'curvature', 'autoencoding', 'denoising', 'normal', 'inpainting', 'segment_unsup2d', 'segment_unsup25d', 'segment_semantic', 'segment_instance']
FEED_FORWARD_TASKS = ['class_object', 'class_scene', 'room_layout', 'vanishing_point']
SINGLE_IMAGE_TASKS = PIX_TO_PIX_TASKS + FEED_FORWARD_TASKS
SIAMESE_TASKS = ['fix_pose', 'jigsaw', 'ego_motion', 'point_match', 'non_fixated_pose']




import torch
import pickle as pkl
import pandas as pd
import matplotlib
import glob
import json
import os
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import torch.nn.functional as F
import torch
from torch.nn.parallel import parallel_apply

def gaussian(M, std, sym=True, device=None):
    if M < 1:
        return torch.tensor([])
    if M == 1:
        return torch.ones((1,))
    odd = M % 2
    if not sym and not odd:
        M = M + 1
    n = torch.arange(0, M, device=device) - (M - 1.0) / 2.0
    sig2 = 2 * std * std
    w = torch.exp(-n ** 2 / sig2)
    if not sym and not odd:
        w = w[:-1]
    return w

def separable_gaussian(img, r=3.5, cutoff=None, device=None):
    if device is None:
        device = img.device
    if r < 1e-1:
        return img
    if cutoff is None:
        cutoff = int(r * 5)
        if (cutoff % 2) == 0: cutoff += 1

    assert (cutoff % 2) == 1
    img = img.to(device)
    _, n_channels, w, h = img.shape
    std = r
    fil = gaussian(cutoff, std, device=device).to(device)
    filsum = fil.sum() #/ n_channels
    fil = torch.stack([fil] * n_channels, dim=0)
    r_pad = int(cutoff) 
    r_pad_half = r_pad // 2

    
    img = F.pad(img, (r_pad_half, r_pad_half, r_pad_half, r_pad_half), "replicate", 0)  # effectively zero padding
    filtered = F.conv2d(img, fil.unsqueeze(1).unsqueeze(-2), bias=None, stride=1, padding=0, dilation=1, groups=n_channels)
    filtered /= filsum
    filtered = F.conv2d(filtered, fil.unsqueeze(1).unsqueeze(-1), bias=None, stride=1, padding=0, dilation=1, groups=n_channels)
    filtered /= filsum

    return filtered

def boxblur(img, r):
    _, _, w, h = img.shape
    f = BoxFilter(r)
    r_pad = int(r * 2)
    r_pad_half = r_pad // 2
    im = F.pad(img, (r_pad_half, r_pad_half, r_pad_half, r_pad_half), "replicate", 0)  # effectively zero padding
    filtered = f(im)
    filtered = filtered[..., r_pad_half:w+r_pad_half, r_pad_half:h+r_pad_half]
    filtered /= ((2 * r + 1) ** 2)
    return filtered


def compute_circle_of_confusion(depths, aperture_size, focal_length, focus_distance):
    assert focus_distance > focal_length
    c = aperture_size * torch.abs(depths - focus_distance) / depths * (focal_length / (focus_distance - focal_length))
    return c


def compute_circle_of_confusion_no_magnification(depths, aperture_size, focus_distance):
    c = aperture_size * torch.abs(depths - focus_distance) / depths 
    return c


def compute_quantiles(depth, quantiles, eps=0.0001):
    depth_flat = depth.reshape(depth.shape[0], -1)
    quantile_vals = torch.quantile(depth_flat, quantiles, dim=1)
    quantile_vals[0] -= eps
    quantile_vals[-1] += eps
    return quantiles, quantile_vals


def compute_quantile_membership(depth, quantile_vals):
    quantile_dists = quantile_vals[1:]  - quantile_vals[:-1]
    depth_flat = depth.reshape(depth.shape[0], -1)
    calculated_quantiles = torch.searchsorted(quantile_vals, depth_flat)
    calculated_quantiles_left = calculated_quantiles - 1
    quantile_vals_unsqueezed = quantile_vals #.unsqueeze(-1).unsqueeze(-1)
    quantile_right = torch.gather(quantile_vals_unsqueezed, 1, calculated_quantiles).reshape(depth.shape)
    quantile_left = torch.gather(quantile_vals_unsqueezed, 1, calculated_quantiles_left).reshape(depth.shape)
    quantile_dists = quantile_right - quantile_left

    dist_right = ((quantile_right - depth) / quantile_dists) #/ quantile_dists[calculated_quantiles_left]
    dist_left = ((depth - quantile_left) / quantile_dists)  #/ quantile_dists[calculated_quantiles_left]
    return dist_left, dist_right, calculated_quantiles_left.reshape(depth.shape), calculated_quantiles.reshape(depth.shape)

def get_blur_stack_single_image(rgb, blur_radii, cutoff_multiplier):
    args = []
    for r in blur_radii:
        cutoff = None if cutoff_multiplier is None else int(r * cutoff_multiplier)
        if cutoff is not None and (cutoff % 2) == 0:
            cutoff += 1
        args.append((rgb, r, cutoff))
    
    blurred_ims = parallel_apply([separable_gaussian]*len(args), args)
    blurred_ims = torch.stack(blurred_ims, dim=1)
    return blurred_ims

def get_blur_stack(rgb, blur_radii, cutoff_multiplier=None):
    args = [(image.unsqueeze(0), radii, cutoff_multiplier) for image, radii in zip(rgb, blur_radii)]
    modules = [get_blur_stack_single_image for _ in args]
    outputs = parallel_apply(modules, args)
    return torch.cat(outputs, dim=0)


def composite_blur_stack(blur_stack, dist_left, dist_right, values_left, values_right):
    shape = list(blur_stack.shape)
    shape[2] = 1
    composite_vals = torch.zeros(shape, dtype=torch.float32, device=blur_stack.device)
    sim_left = (1 - dist_left**2)
    sim_right = (1 - dist_right**2)

    if len(composite_vals.shape) == len(values_left.unsqueeze(1).shape):
        _ = composite_vals.scatter_(1, index=values_left.unsqueeze(1), src=sim_left.unsqueeze(1))
        _ = composite_vals.scatter_(1, index=values_right.unsqueeze(1), src=sim_right.unsqueeze(1))
    else:
        _ = composite_vals.scatter_(1, index=values_left.unsqueeze(1).unsqueeze(2), src=sim_left.unsqueeze(1).unsqueeze(2))
        _ = composite_vals.scatter_(1, index=values_right.unsqueeze(1).unsqueeze(2), src=sim_right.unsqueeze(1).unsqueeze(2))

    composite_vals /= composite_vals.sum(dim=1, keepdims=True)
    composited = composite_vals * blur_stack
    composited = composited.sum(dim=1)
    return composited


def refocus_image(rgb, depth, focus_distance, aperture_size, quantile_vals, return_segments=False):
    quantile_vals_squeezed = quantile_vals.squeeze()
    dist_left, dist_right, calculated_quantiles_left, calculated_quantiles = compute_quantile_membership(depth, quantile_vals)
    blur_radii = compute_circle_of_confusion_no_magnification(quantile_vals, aperture_size, focus_distance)

    # print(f'Quantiles: {quantile_vals}')
    # print(f'Blurs: {blur_radii.shape}')

    blur_stack = get_blur_stack(rgb, blur_radii, cutoff_multiplier=3)
    composited = composite_blur_stack(blur_stack, dist_left, dist_right, calculated_quantiles_left, calculated_quantiles)

    if return_segments:
        return composited, calculated_quantiles_left
    else:
        return composited

##########################################
#
##########################################

def RefocusImageAugmentation(n_quantiles, aperture_min, aperture_max, return_segments=False):
    '''
        RefocusImageAugmentation
            Given an RGB image and depth image, refocuses the RGB image by blurring according to depth.
            - Based on a paraxial thin-lens model for refocusing.
            - Bands the depth image into segments of equal size (size given by n_quantiles)
            - Finds blur corresponding to each depth segment cutoff
            - Creates "blur stack" of n_quantiles + 1 blurred image (separable Gaussian filter)
            - For each depth location, composite pixel value as interpolation between two of the 
                blur images from the stack           
        
            Randomly selects:
            - focus distance to be one of the quantiles (so that some of the image is always in focus).
            - aperture_size log-uniformly from [aperture_min, aperture_max]

        Arguments:
            n_quantiles: How many quantiles to use in the blur stack. More is better, but slower
            aperture_min: Smallest aperture to use
            aperture_max: Largest aperture to use (randomly selects one log-uniformly in this range)
            return_segments: whether to return the depth segments
    '''
    def refocus_image_(rgb, depth):
        with torch.no_grad():
            device = depth.device
            quantiles = torch.arange(0, n_quantiles + 1, device=device) / n_quantiles
            quantiles, quantile_vals = compute_quantiles(depth, quantiles, eps=0.0001)
            quantile_vals = quantile_vals.permute(1, 0)

            focus_dist_idxs = torch.randint(low=1, high=n_quantiles, size=(rgb.shape[0],), device=device)
            focus_dists = torch.gather(quantile_vals, 1, focus_dist_idxs.unsqueeze(1))#.permute(1,0)

            log_min = torch.log(torch.tensor(aperture_min, device=device))
            log_max = torch.log(torch.tensor(aperture_max, device=device))

            apertures = torch.exp(
                            torch.rand(size=(rgb.shape[0],1), device=device) 
                            * (log_max - log_min) + log_min
                        )
            # print(f'dist: {focus_dists}, apertures: {apertures}')
            return refocus_image(rgb, depth, focus_dists, apertures, quantile_vals, return_segments)
    return refocus_image_



import torch
from torch import nn


class ECALayer(nn.Module):
    """Constructs a ECA module.
    Args:
        channel: Number of channels of the input feature map
        k_size: Adaptive selection of kernel size
    """
    def __init__(self, channel, k_size=3):
        super(ECALayer, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.conv = nn.Conv1d(1, 1, kernel_size=k_size, padding=(k_size - 1) // 2, bias=False)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        # x: input features with shape [b, c, h, w]
        b, c, h, w = x.size()

        # feature descriptor on the global spatial information
        y = self.avg_pool(x)

        # Two different branches of ECA module
        y = self.conv(y.squeeze(-1).transpose(-1, -2)).transpose(-1, -2).unsqueeze(-1)

        # Multi-scale information fusion
        y = self.sigmoid(y)

        return x * y.expand_as(x)


class ChannelAttention(nn.Module):
    def __init__(self, num_features, reduction):
        super(ChannelAttention, self).__init__()
        self.module = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(num_features, num_features // reduction, kernel_size=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(num_features // reduction, num_features, kernel_size=1),
            nn.Sigmoid()
        )

    def forward(self, x):
        return x * self.module(x)


class RCAB(nn.Module):
    def __init__(self, num_features, reduction):
        super(RCAB, self).__init__()
        self.module = nn.Sequential(
            nn.Conv2d(num_features, num_features, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(num_features, num_features, kernel_size=3, padding=1),
            ChannelAttention(num_features, reduction)
        )

    def forward(self, x):
        return x + self.module(x)


class RG(nn.Module):
    def __init__(self, num_features, num_rcab, reduction):
        super(RG, self).__init__()
        self.module = [RCAB(num_features, reduction) for _ in range(num_rcab)]
        self.module.append(nn.Conv2d(num_features, num_features, kernel_size=3, padding=1))
        self.module = nn.Sequential(*self.module)

    def forward(self, x):
        return x + self.module(x)


class RCAN(nn.Module):
    def __init__(self, scale, num_features, num_rg, num_rcab, reduction):
        super(RCAN, self).__init__()
        self.sf = nn.Conv2d(3, num_features, kernel_size=3, padding=1)
        self.rgs = nn.Sequential(*[RG(num_features, num_rcab, reduction) for _ in range(num_rg)])
        self.conv1 = nn.Conv2d(num_features, num_features, kernel_size=3, padding=1)
        self.upscale = nn.Sequential(
            nn.Conv2d(num_features, num_features * (scale ** 2), kernel_size=3, padding=1),
            nn.PixelShuffle(scale)
        )
        self.conv2 = nn.Conv2d(num_features, 3, kernel_size=3, padding=1)

    def forward(self, x):
        x = self.sf(x)
        residual = x
        x = self.rgs(x)
        x = self.conv1(x)
        x += residual
        x = self.upscale(x)
        x = self.conv2(x)
        return x


class CBAMChannelAttention(nn.Module):
    def __init__(self, in_planes, ratio=16):
        super(CBAMChannelAttention, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.max_pool = nn.AdaptiveMaxPool2d(1)

        self.fc1   = nn.Conv2d(in_planes, in_planes // 16, 1, bias=False)
        self.relu1 = nn.ReLU()
        self.fc2   = nn.Conv2d(in_planes // 16, in_planes, 1, bias=False)

        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avg_out = self.fc2(self.relu1(self.fc1(self.avg_pool(x))))
        max_out = self.fc2(self.relu1(self.fc1(self.max_pool(x))))
        out = avg_out + max_out
        return self.sigmoid(out)


class CBAMSpatialAttention(nn.Module):
    def __init__(self, kernel_size=7):
        super(CBAMSpatialAttention, self).__init__()

        assert kernel_size in (3, 7), 'kernel size must be 3 or 7'
        padding = 3 if kernel_size == 7 else 1

        self.conv1 = nn.Conv2d(2, 1, kernel_size, padding=padding, bias=False)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avg_out = torch.mean(x, dim=1, keepdim=True)
        max_out, _ = torch.max(x, dim=1, keepdim=True)
        x = torch.cat([avg_out, max_out], dim=1)
        x = self.conv1(x)
        return self.sigmoid(x)



import torch
import torch.nn.functional as F
from torchvision import transforms

import PIL
from PIL import Image
import numpy as np
import matplotlib.pyplot as plt

import argparse
import os.path
from pathlib import Path
import glob
import sys
import pdb

from data.refocus_augmentation import RefocusImageAugmentation
from data.transforms import get_transform

parser = argparse.ArgumentParser(description='Visualize 3D refocus augmentation')

parser.add_argument('--num_quantiles', dest='num_quantiles', \
    help="number of qualtiles to use in blur stack. More is better, but slower", required=False)
parser.set_defaults(num_quantiles=10)

parser.add_argument('--min_aperture', dest='min_aperture', help="smallest aperture to use", required=False)
parser.set_defaults(min_aperture=0.001)

parser.add_argument('--max_aperture', dest='max_aperture', help="largest aperture to use", required=False)
parser.set_defaults(max_aperture=6)

parser.add_argument('--input_path', dest='input_path', \
    help="path to folder containing rgb and depth_euclidean")
parser.set_defaults(im_name='NONE')

parser.add_argument('--output_path', dest='output_path',\
    help="path to where refocused rgb should be stored")
    
parser.set_defaults(store_name='NONE')

args = parser.parse_args()

transform_rgb = get_transform('rgb', image_size=512)
transform_depth = get_transform('depth_euclidean', image_size=512)
trans_topil = transforms.ToPILImage()

os.system(f"mkdir -p {args.output_path}")
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

refocus_aug = RefocusImageAugmentation(
    args.num_quantiles, float(args.min_aperture), float(args.max_aperture), return_segments=False)


def save_outputs(img_path, output_file_name):
    if not output_file_name.__contains__('rgb'): return
    rgb_path = img_path
    depth_path = img_path.replace('rgb', 'depth_euclidean')
    save_path = args.output_path+'/'+output_file_name+'_refocused'+'.png'

    print(f'Reading input {img_path} ...')
    rgb = Image.open(rgb_path)
    rgb_tensor = transform_rgb(rgb)[:3].unsqueeze(0).to(device)
    depth = Image.open(depth_path)
    depth_tensor = transform_depth(depth)[:1].unsqueeze(0).to(device)

    if rgb_tensor.shape[1] == 1:
        rgb_tensor = rgb_tensor.repeat_interleave(3,1)

    augmented_rgb = refocus_aug(rgb_tensor, depth_tensor)

    print(f'Writing output {save_path} ...')
    trans_topil(augmented_rgb[0]).save(save_path)


input_path = Path(args.input_path)

if input_path.is_dir():
    for f in glob.glob(args.input_path+'/*'):
        save_outputs(f, os.path.splitext(os.path.basename(f))[0])
else:
    print("invalid file path!")
    sys.exit()



import functools, requests
import multiprocessing as mp
from ..metadata import RemoteBucketStorageMetadata, ZippedModel


# SPLITS | Which spaces to download

from .component_datasets import taskonomy, hypersim, replica, replica_gso, blendedmvg

# --- Specifics ---
class OmnidataMetadata(RemoteBucketStorageMetadata):
  def __init__(self, base_url='https://datasets.epfl.ch/omnidata/', **kwargs):
    super().__init__(base_url=base_url, **kwargs)

  @functools.cached_property
  def links(self): return [k for k in requests.get(self.link_file).text.splitlines()
      if k.endswith(self.expected_suffix) and not any([d in k for d in ('depth_zbuffer2', 'mask_valid2')])]
  
  def parse(self, url) -> ZippedModel:
    '''Most omnidata files are stored like: https://datasets.epfl.ch/omnidata//omnidata_tars/depth_euclidean/blendedMVS/depth_euclidean-blendedMVS-000000000000000000000000.tar.gz'''
    split_urls = url.split('/') 
    if not ((url_len := len(split_urls)) == 8): raise ValueError(f'Expected url to be split into 8 components, not {url_len}: "{url}"')
    domain, component_name, fname = split_urls[5:8]
    domain2, component_name2, *model_name = fname[:-len(self.expected_suffix)].split('-')
    if not url.endswith(self.expected_suffix): raise ValueError(f'Expected compressed url to end with "{self.expected_suffix}": {url}')
    if not component_name == component_name2: raise ValueError(f'Expected domain from fname ("{component_name2}" in "{fname}") to match domain in url ("{component_name}" in "{url}").') 
    if not domain == domain2: raise ValueError(f'Expected domain from fname ("{domain2}" in "{fname}") to match domain in url ("{domain}" in "{url}").') 
    if len(model_name) == 0:  raise ValueError(f'Model name in {fname} has 0 length after domain + component.')  
    model_name = "-".join(model_name)
    return ZippedModel(component_name=component_name, domain=domain, model_name=model_name, url=url, tar_structure=self.tar_structure, checksum=self.checksum(url))


class TaskonomyMetadata(RemoteBucketStorageMetadata):
  def __init__(self, base_url='https://datasets.epfl.ch/taskonomy/', expected_suffix='.tar', tar_structure=('domain',), **kwargs):
    super().__init__(base_url=base_url, expected_suffix=expected_suffix, tar_structure=tar_structure, **kwargs)

  def parse(self, url) -> ZippedModel:
    '''Taskonony links are stored as: https://datasets.epfl.ch/taskonomy/adairsville_class_object.tar'''
    split_urls = url.split('/') 
    if not url.endswith(self.expected_suffix): raise ValueError(f'Expected compressed url to end with "{self.expected_suffix}": {url}')
    if not (url_len := len(split_urls)) == 5: raise ValueError(f'Expected url to be split into 5 components, not {url_len}: "{url}"')
    if not (component_name := split_urls[-2]) == 'taskonomy': raise ValueError(f'Expected component name to be "taskonomy", not "{component_name}"')
    fname = split_urls[-1][:-len(self.expected_suffix)]
    model_name, *domain = fname.split('_')
    if len(domain) == 0:  raise ValueError(f'Domain name in {fname} has 0 length after model.')  
    domain = "_".join(domain)
    return ZippedModel(component_name=component_name, domain=domain, model_name=model_name, url=url, tar_structure=self.tar_structure, checksum=self.checksum(url))

STARTER_DATASET_REMOTE_SERVER_METADATAS = [
  OmnidataMetadata( base_url='https://datasets.epfl.ch/omnidata/', expected_suffix='.tar'),
  TaskonomyMetadata( base_url='https://datasets.epfl.ch/taskonomy/'),
]


STARTER_DATA_COMPONENTS = {
  'taskonomy': taskonomy,
  'hypersim': hypersim,
  'replica': replica,
  'replica_gso': replica_gso,
  'blendedmvg': blendedmvg,
  'hm3d': None,
  'clevr_simple': None,
  'clevr_complex': None
}

STARTER_DATA_COMPONENT_TO_SPLIT = {k: getattr(v, 'split_to_spaces', None) for k, v in STARTER_DATA_COMPONENTS.items()}
STARTER_DATA_COMPONENT_TO_SUBSET = {k: getattr(v, 'subset_to_spaces', None) for k, v in STARTER_DATA_COMPONENTS.items()}


STARTER_DATA_LICENSES = {
  'omnidata':       'https://raw.githubusercontent.com/EPFL-VILAB/omnidata-tools/main/LICENSE',
  'replica':        'https://raw.githubusercontent.com/facebookresearch/Replica-Dataset/main/LICENSE',
  'hypersim':       'https://raw.githubusercontent.com/apple/ml-hypersim/master/LICENSE.txt',
  'replica_gso': 'https://creativecommons.org/licenses/by/4.0/legalcode',
  'clevr_simple':   'https://creativecommons.org/licenses/by/4.0/legalcode',
  'clevr_complex':  'https://creativecommons.org/licenses/by/4.0/legalcode',
  'blendedmvg':     'https://creativecommons.org/licenses/by/4.0/legalcode',
  'taskonomy':      'https://raw.githubusercontent.com/StanfordVL/taskonomy/master/data/LICENSE',
  'hm3d':           'https://matterport.com/matterport-end-user-license-agreement-academic-use-model-data'
}

if __name__ == '__main__':
  print(STARTER_DATA_COMPONENT_TO_SUBSET['replica'])
  [print(rsm.info) for rsm in STARTER_DATASET_REMOTE_SERVER_METADATAS]



import functools, operator 

import torch
import torch.nn.functional as F
import warnings

from .task_configs import task_parameters as TASK_PARAMETERS

DEFAULT_MASK_EXTRA_RADIUS=5

def make_mask(tensor, task):
    return build_mask(tensor.unsqueeze(1), val=task_parameters[task]['mask_val'])[0]

def build_mask(target, val=0.0, tol=1e-3, mask_extra_radius=DEFAULT_MASK_EXTRA_RADIUS):
    if target.shape[1] == 1:
        mask = ((target >= val - tol) & (target <= val + tol))
        #mask = F.conv2d(mask.float(), torch.ones(1, 1, 5, 5, device=mask.device), padding=2) != 0
        mask = (0 != F.conv2d(mask.float(),
                        torch.ones(1, 1, mask_extra_radius, mask_extra_radius, device=mask.device),
                        padding=(mask_extra_radius // 2)) )
        return (~mask).expand_as(target)


    masks = [target[:, t] for t in range(target.shape[1])]
    masks = [(t >= val - tol) & (t <= val + tol) for t in masks]
    mask = functools.reduce(lambda a,b: a&b, masks).unsqueeze(1)
    mask = (0 != F.conv2d(mask.float(),
                        torch.ones(1, 1, mask_extra_radius, mask_extra_radius, device=mask.device),
                        padding=(mask_extra_radius // 2)) )
#     mask = F.conv2d(mask.float(), torch.ones(1, 1, 5, 5, device=mask.device), padding=2) != 0
    return (~mask).expand_as(target)


def make_mask_from_data(tensors, tasks, mask_extra_radius=DEFAULT_MASK_EXTRA_RADIUS, device='cpu'):
    ''' Makes mask from a list of tensors (and their associated tasks)
        Args:
            tensors: Iterable of torch.tensors
            tasks: List of tasks, named as in .task_configs.py
        Returns:
            Mask to be used for all the image tasks in `tensors`
    '''
    task_to_tensor = {task: tensor for task, tensor in zip(tasks, tensors)}
    if 'mask_valid' in tasks: # resizing might mess with exact values, so prioritize mask_val.
        return build_mask(task_to_tensor['mask_valid'],
                          mask_extra_radius=mask_extra_radius,
                          val=TASK_PARAMETERS['mask_valid']['mask_val']) 

    # could be improved by priotizing images w/fewer channels
    for tensor, task in zip(tensors, tasks):
        if 'mask_val' in TASK_PARAMETERS[task]:
            tensor_size = tensor.shape[-1]
            if tensor_size != 512 and task != 'mask_valid':
                warnings.warn(f"Making mask from task \"{task}\" with size {tensor_size}, which has been resized from 512. Masks may be inaccurate.", RuntimeWarning)
            return build_mask(tensor,
                              mask_extra_radius=mask_extra_radius,
                              val=TASK_PARAMETERS[task]['mask_val'])

    raise ValueError(f'Could not make mask for any task in {tasks}')
    


import csv
import os

def get_splits(split_path, forbidden_buildings=[]):
    with open(split_path) as csvfile:
        readCSV = csv.reader(csvfile, delimiter=',')
        train_list, val_list, test_list = [], [], []
        for row in readCSV:
            name, is_train, is_val, is_test = row
            if name in forbidden_buildings: continue
            if is_train == '1': train_list.append(name)
            if is_val == '1':   val_list.append(name)
            if is_test == '1':  test_list.append(name)
    return {
        'train': sorted(train_list),
        'val': sorted(val_list),
        'test': sorted(test_list)
    }

def get_all_spaces(splits): return set(sum(splits.values(), start=[]))


import torch
import torch.nn.functional as F
from torchvision import transforms

import PIL
from PIL import Image
import numpy as np
import matplotlib.pyplot as plt

import argparse
import os.path
from pathlib import Path
import glob
import sys

import pdb

from modules.unet import UNet
from modules.midas.dpt_depth import DPTDepthModel
from data.transforms import get_transform


def depth_to_heatmap(img):   
    img = (img - np.min(img)) / np.ptp(img)
    cm = plt.get_cmap('viridis', 2**16)
    pixel_colored = np.uint8(np.rint(cm(img) * 255))[:, :, :3]
    return pixel_colored

parser = argparse.ArgumentParser(description='Visualize output for depth or surface normals')

parser.add_argument('--task', dest='task', help="normal or depth")
parser.set_defaults(task='NONE')

parser.add_argument('--img_path', dest='img_path', help="path to rgb image")
parser.set_defaults(im_name='NONE')

parser.add_argument('--output_path', dest='output_path', help="path to where output image should be stored")
parser.set_defaults(store_name='NONE')

args = parser.parse_args()

root_dir = './pretrained_models/'

trans_topil = transforms.ToPILImage()

os.system(f"mkdir -p {args.output_path}")
map_location = (lambda storage, loc: storage.cuda()) if torch.cuda.is_available() else torch.device('cpu')
device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')


# get target task and model
if args.task == 'normal':
    image_size = 512
    pretrained_weights_path = root_dir + 'omnidata_rgb2normal_unet.pth'
    model = UNet(in_channels=3, out_channels=3)
    checkpoint = torch.load(pretrained_weights_path, map_location=map_location)
    if 'state_dict' in checkpoint:
        state_dict = {}
        for k, v in checkpoint['state_dict'].items():
            state_dict[k.replace('model.', '')] = v
    else:
        state_dict = checkpoint
    model.load_state_dict(state_dict)
    model.to(device)
    trans_totensor = transforms.Compose([transforms.Resize(image_size, interpolation=PIL.Image.BILINEAR),
                                        transforms.CenterCrop(image_size),
                                        get_transform('rgb', image_size=None)])

elif args.task == 'depth':
    image_size = 384
    pretrained_weights_path = root_dir + 'omnidata_rgb2depth_dpt_hybrid.pth'
    # model = DPTDepthModel(backbone='vitl16_384') # DPT Large
    model = DPTDepthModel(backbone='vitb_rn50_384') # DPT Hybrid
    checkpoint = torch.load(pretrained_weights_path, map_location=map_location)
    if 'state_dict' in checkpoint:
        state_dict = {}
        for k, v in checkpoint['state_dict'].items():
            state_dict[k[6:]] = v
    else:
        state_dict = checkpoint
    model.load_state_dict(state_dict)
    model.to(device)
    trans_totensor = transforms.Compose([transforms.Resize(image_size, interpolation=PIL.Image.BILINEAR),
                                        transforms.CenterCrop(image_size),
                                        transforms.ToTensor(),
                                        transforms.Normalize(mean=0.5, std=0.5)])

else:
    print("task should be one of the following: normal, depth")
    sys.exit()

trans_rgb = transforms.Compose([transforms.Resize(512, interpolation=PIL.Image.BILINEAR),
                                transforms.CenterCrop(512)])


def save_outputs(img_path, output_file_name):
    save_path = os.path.join(args.output_path, f'{output_file_name}_{args.task}.png')

    print(f'Reading input {img_path} ...')
    img = Image.open(img_path)

    img_tensor = trans_totensor(img)[:3].unsqueeze(0).to(device)

    rgb_path = os.path.join(args.output_path, f'{output_file_name}_rgb.png')
    trans_rgb(img).save(rgb_path)

    if img_tensor.shape[1] == 1:
        img_tensor = img_tensor.repeat_interleave(3,1)

    output = model(img_tensor).clamp(min=0, max=1)

    if args.task == 'depth':
        output = F.interpolate(output.unsqueeze(0), (512, 512), mode='bicubic').squeeze(0)
        output = 1 / (output + 1e-6)
        output = torch.tensor(depth_to_heatmap(output.detach().cpu().squeeze().numpy())).permute(2,0,1).unsqueeze(0)
        output = (output - output.min()) / (output.max() - output.min())

    trans_topil(output[0]).save(save_path)
    print(f'Writing output {save_path} ...')


img_path = Path(args.img_path)
if img_path.is_file():
    save_outputs(args.img_path, os.path.splitext(os.path.basename(args.img_path))[0])
elif img_path.is_dir():
    for f in glob.glob(args.img_path+'/*'):
        save_outputs(f, os.path.splitext(os.path.basename(f))[0])
else:
    print("invalid file path!")
    sys.exit()


import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from .channel_attention import *


class UNet_up_block(nn.Module):
    def __init__(self, prev_channel, input_channel, output_channel, up_sample=True, use_skip=True):
        super().__init__()
        self.up_sampling = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=False)
        if use_skip:
            self.conv1 = nn.Conv2d(prev_channel + input_channel, output_channel, 3, padding=1)
        else:
            self.conv1 = nn.Conv2d(input_channel, output_channel, 3, padding=1)
        self.bn1 = nn.GroupNorm(8, output_channel)
        self.conv2 = nn.Conv2d(output_channel, output_channel, 3, padding=1)
        self.bn2 = nn.GroupNorm(8, output_channel)
        self.conv3 = nn.Conv2d(output_channel, output_channel, 3, padding=1)
        self.bn3 = nn.GroupNorm(8, output_channel)
        self.relu = torch.nn.ReLU()
        self.up_sample = up_sample

    def forward(self, x, prev_feature_map=None):
        if self.up_sample:
            x = self.up_sampling(x)
        if prev_feature_map is not None:
            x = torch.cat((x, prev_feature_map), dim=1)
        x = self.relu(self.bn1(self.conv1(x)))
        x = self.relu(self.bn2(self.conv2(x)))
        x = self.relu(self.bn3(self.conv3(x)))
        return x
 

class UNet_down_block(nn.Module):
    def __init__(self, input_channel, output_channel, down_size=True):
        super().__init__()
        self.conv1 = nn.Conv2d(input_channel, output_channel, 3, padding=1)
        self.bn1 = nn.GroupNorm(8, output_channel)
        self.conv2 = nn.Conv2d(output_channel, output_channel, 3, padding=1)
        self.bn2 = nn.GroupNorm(8, output_channel)
        self.conv3 = nn.Conv2d(output_channel, output_channel, 3, padding=1)
        self.bn3 = nn.GroupNorm(8, output_channel)
        self.max_pool = nn.MaxPool2d(2, 2)
        self.relu = nn.ReLU()
        self.down_size = down_size

    def forward(self, x):
        x = self.relu(self.bn1(self.conv1(x)))
        x = self.relu(self.bn2(self.conv2(x)))
        x = self.relu(self.bn3(self.conv3(x)))
        if self.down_size:
            x = self.max_pool(x)
        return x


class UNet(nn.Module):
    def __init__(self, downsample=6, in_channels=3, out_channels=3, patch_size=1):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.downsample = downsample
        self.patch_size = patch_size

        self.down1 = UNet_down_block(in_channels, 16, False)
        self.down_blocks = nn.ModuleList(
            [UNet_down_block(2**(4+i), 2**(5+i), True) for i in range(0, downsample)]
        )

        bottleneck = 2**(4 + downsample)
        self.mid_conv1 = nn.Conv2d(bottleneck, bottleneck, 3, padding=1)
        self.bn1 = nn.GroupNorm(8, bottleneck)
        self.mid_conv2 = nn.Conv2d(bottleneck, bottleneck, 3, padding=1)
        self.bn2 = nn.GroupNorm(8, bottleneck)
        self.mid_conv3 = torch.nn.Conv2d(bottleneck, bottleneck, 3, padding=1)
        self.bn3 = nn.GroupNorm(8, bottleneck)

        self.up_blocks = nn.ModuleList(
            [UNet_up_block(2**(4+i), 2**(5+i), 2**(4+i)) for i in range(0, downsample)]
        )

        self.last_conv1 = nn.Conv2d(16, 16, 3, padding=1)
        self.last_bn = nn.GroupNorm(8, 16)
        self.last_conv2 = nn.Conv2d(16, out_channels, 1, padding=0)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.down1(x)
        xvals = [x]
        for i in range(0, self.downsample):
            x = self.down_blocks[i](x)
            xvals.append(x)

        x = self.relu(self.bn1(self.mid_conv1(x)))
        x = self.relu(self.bn2(self.mid_conv2(x)))
        x = self.relu(self.bn3(self.mid_conv3(x)))

        for i in range(0, self.downsample)[::-1]:
            x = self.up_blocks[i](x, xvals[i])

        x = self.relu(self.last_bn(self.last_conv1(x)))
        x = self.last_conv2(x)
        # x = F.interpolate(x, scale_factor=(1/self.patch_size, 1/self.patch_size), mode='bilinear', align_corners=False)
        return x




class UNetRelu(nn.Module):
    def __init__(self, downsample=6, in_channels=3, out_channels=3, patch_size=1):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.downsample = downsample
        self.patch_size = patch_size

        self.down1 = UNet_down_block(in_channels, 16, False)
        self.down_blocks = nn.ModuleList(
            [UNet_down_block(2**(4+i), 2**(5+i), True) for i in range(0, downsample)]
        )

        bottleneck = 2**(4 + downsample)
        self.mid_conv1 = nn.Conv2d(bottleneck, bottleneck, 3, padding=1)
        self.bn1 = nn.GroupNorm(8, bottleneck)
        self.mid_conv2 = nn.Conv2d(bottleneck, bottleneck, 3, padding=1)
        self.bn2 = nn.GroupNorm(8, bottleneck)
        self.mid_conv3 = torch.nn.Conv2d(bottleneck, bottleneck, 3, padding=1)
        self.bn3 = nn.GroupNorm(8, bottleneck)

        self.up_blocks = nn.ModuleList(
            [UNet_up_block(2**(4+i), 2**(5+i), 2**(4+i)) for i in range(0, downsample)]
        )

        self.last_conv1 = nn.Conv2d(16, 16, 3, padding=1)
        self.last_bn = nn.GroupNorm(8, 16)
        self.last_conv2 = nn.Conv2d(16, out_channels, 1, padding=0)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.down1(x)
        xvals = [x]
        for i in range(0, self.downsample):
            x = self.down_blocks[i](x)
            xvals.append(x)

        x = self.relu(self.bn1(self.mid_conv1(x)))
        x = self.relu(self.bn2(self.mid_conv2(x)))
        x = self.relu(self.bn3(self.mid_conv3(x)))

        for i in range(0, self.downsample)[::-1]:
            x = self.up_blocks[i](x, xvals[i])

        x = self.relu(self.last_bn(self.last_conv1(x)))
        x = self.last_conv2(x)
        # x = F.interpolate(x, scale_factor=(1/self.patch_size, 1/self.patch_size), mode='bilinear', align_corners=False)
        return self.relu(x)


class UNetV2(nn.Module):
    def __init__(self, in_channels=3, out_channels=3, patch_size=1):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.patch_size = patch_size

        self.down1 = UNet_down_block(in_channels, 16, False)
        self.down_blocks = nn.ModuleList([
            UNet_down_block(16, 32, True),
            UNet_down_block(32, 64, True),
            UNet_down_block(64, 256, True),
            UNet_down_block(256, 256, True),
            UNet_down_block(256, 512, True),
            UNet_down_block(512, 1024, True),
        ])

        bottleneck = 1024
        self.mid_conv1 = nn.Conv2d(bottleneck, bottleneck, 3, padding=1)
        self.bn1 = nn.GroupNorm(8, bottleneck)
        self.mid_conv2 = nn.Conv2d(bottleneck, bottleneck, 3, padding=1)
        self.bn2 = nn.GroupNorm(8, bottleneck)
        self.mid_conv3 = torch.nn.Conv2d(bottleneck, bottleneck, 3, padding=1)
        self.bn3 = nn.GroupNorm(8, bottleneck)

        self.up_blocks = nn.ModuleList([
            UNet_up_block(512, 1024, 512),
            UNet_up_block(256, 512, 256),
            UNet_up_block(256, 256, 256),
            UNet_up_block(64, 256, 64),
            UNet_up_block(32, 64, 32),
            UNet_up_block(16, 32, 16),
        ])

        self.last_conv1 = nn.Conv2d(16, 16, 3, padding=1)
        self.last_bn = nn.InstanceNorm2d(16)
        self.last_conv2 = nn.Conv2d(16, out_channels, 1, padding=0)
        self.attention = ECALayer(out_channels, k_size=7)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.down1(x)
        xvals = [x]
        for down_block in self.down_blocks:
            x = down_block(x)
            xvals.append(x)

        x = self.relu(self.bn1(self.mid_conv1(x)))
        x = self.relu(self.bn2(self.mid_conv2(x)))
        x = self.relu(self.bn3(self.mid_conv3(x)))

        for up_block, xval in zip(self.up_blocks, xvals[::-1][1:len(self.up_blocks)+1]):
            x = up_block(x, xval)

        x = self.relu(self.last_bn(self.last_conv1(x)))
        x = self.last_conv2(x)
        x = self.attention(x)
        x = F.interpolate(x, scale_factor=(1/self.patch_size, 1/self.patch_size), mode='bilinear', align_corners=False)
        return x




import json
import numpy as np
import os
from PIL import Image
import h5py
import torch
from scipy.ndimage.filters import convolve
import torchvision
import torchvision.transforms as transforms
import torch.nn as nn
import torch.nn.functional as F
from   typing import Optional

from . import task_configs

try:
    import accimage
except ImportError:
    pass

# from tlkit.utils import TASKS_TO_CHANNELS, FEED_FORWARD_TASKS

MAKE_RESCALE_0_1_NEG1_POS1   = lambda n_chan: transforms.Normalize([0.5]*n_chan, [0.5]*n_chan)
RESCALE_0_1_NEG1_POS1        = transforms.Normalize([0.5], [0.5])  # This needs to be different depending on num out chans
MAKE_RESCALE_0_MAX_NEG1_POS1 = lambda maxx: transforms.Normalize([maxx / 2.], [maxx * 1.0])
RESCALE_0_255_NEG1_POS1      = transforms.Normalize([127.5,127.5,127.5], [255, 255, 255])
MAKE_RESCALE_0_MAX_0_POS1 = lambda maxx: transforms.Normalize([0.0], [maxx * 1.0])

def get_transform(task: str, image_size=Optional[int]):
   
    if task in ['rgb', 'normal', 'reshading']:
        transform = transform_8bit
    elif task in ['mask_valid']:
        transform = transforms.ToTensor()
    elif task in ['keypoints2d', 'keypoints3d', 'depth_euclidean', 'depth_zbuffer', 'edge_texture', 'edge_occlusion']:
#         return transform_16bit_int
        transform = transform_16bit_single_channel
    elif task in ['principal_curvature', 'curvature']:
        transform = transform_8bit_n_channel(2)
    elif task in ['segment_semantic', 'segment_instance', 'fragments']:  # this is stored as 1 channel image (H,W) where each pixel value is a different class
        transform = transform_dense_labels
#     elif len([t for t in FEED_FORWARD_TASKS if t in task]) > 0:
#         return torch.Tensor
#     elif 'decoding' in task:
#         return transform_16bit_n_channel(TASKS_TO_CHANNELS[task.replace('_decoding', '')])
#     elif 'encoding' in task:
#         return torch.Tensor
    elif task in ['class_object', 'class_scene']:
        transform = torch.Tensor
        image_size = None
    elif task in ['segment_panoptic']:
        # transform = transform_8bit_n_channel(n_channel=1, crop_channels=True)
        transform = transform_dense_labels
    elif task in ['mesh', 'point_info']:
        return None
    else:
        raise NotImplementedError("Unknown transform for task {}".format(task))
    
    if 'clamp_to' in task_configs.task_parameters[task]:
        minn, maxx = task_configs.task_parameters[task]['clamp_to']
        if minn > 0:
            raise NotImplementedError("Rescaling (min1, max1) -> (min2, max2) not implemented for min1, min2 != 0 (task {})".format(task))
        transform = transforms.Compose([
                        transform,
                        MAKE_RESCALE_0_MAX_0_POS1(maxx)])

    if image_size is not None:
        if task == 'fragments':
            resize_frag = lambda frag: torch.nn.functional.interpolate(frag.permute(2,0,1).unsqueeze(0).float(), image_size, mode='nearest').long()[0].permute(1,2,0)
            transform = transforms.Compose([
                transform,
                resize_frag
            ])
        else:
            # resize_method = Image.BILINEAR if task not in ['segment_panoptic', 'segment_instance', 'segment_semantic', 'mask_valid'] else Image.NEAREST
            resize_method = Image.BILINEAR if task in ['rgb'] else Image.NEAREST
            transform = transforms.Compose([
                transforms.Resize(image_size, resize_method),
                transform])

    return transform

# For semantic segmentation
transform_dense_labels = lambda img: torch.Tensor(np.array(img)).long()  # avoids normalizing

# Transforms to a 3-channel tensor and then changes [0,1] -> [-1, 1]
transform_8bit = transforms.Compose([
        transforms.ToTensor(),
#         MAKE_RESCALE_0_1_NEG1_POS1(3),
    ])
    
# Transforms to a n-channel tensor and then changes [0,1] -> [-1, 1]. Keeps only the first n-channels
def transform_8bit_n_channel(n_channel=1, crop_channels=False):
    if crop_channels:
        crop_channels_fn = lambda x: x[:n_channel] if x.shape[0] > n_channel else x
    else: 
        crop_channels_fn = lambda x: x
    return transforms.Compose([
            transforms.ToTensor(),
            crop_channels_fn,
#             MAKE_RESCALE_0_1_NEG1_POS1(n_channel),
        ])

# Transforms to a 1-channel tensor and then changes [0,1] -> [-1, 1].
def transform_16bit_single_channel(im):
    im = transforms.ToTensor()(im)
    im = im.float() / (2 ** 16 - 1.0) 
#     return RESCALE_0_1_NEG1_POS1(im)
    return im


def transform_16bit_n_channel(n_channel=1):
    if n_channel == 1:
        return transform_16bit_single_channel # PyTorch handles these differently
    else:
        return transforms.Compose([
            transforms.ToTensor(),
#             MAKE_RESCALE_0_1_NEG1_POS1(n_channel),
        ])



def default_loader(path):
    if '.hdf5' in path:  # semantic labels for hypersim
        with h5py.File(path, "r") as f:
            data=f['dataset'][:]
            data_arr = np.repeat(np.expand_dims(data, axis=2), 3, axis=2)
            return Image.fromarray(np.uint8(data_arr))
    elif '.npy' in path:
        return np.load(path)
    elif '.json' in path:
        with open(path, 'r') as f:
            data_dict = json.load(f)
            data_dict['building'] = os.path.basename(os.path.dirname(path))
            data_dict.pop('nonfixated_points_in_view')
#             new_data = {}
#             new_data['camera_location'] = data_dict['camera_location']
#             new_data['camera_rotation_final'] = data_dict['camera_rotation_final']
#             new_data['field_of_view_rads'] = data_dict['field_of_view_rads']
#             data_dict = new_data
            return data_dict
    else:
        from torchvision import get_image_backend
        if get_image_backend() == 'accimage':
            im = accimage_loader(path)
        else:
            im = pil_loader(path)
        return im

def pil_loader(path):
    # open path as file to avoid ResourceWarning (https://github.com/python-pillow/Pillow/issues/835)
    # path = path.replace('/scratch/', '/scratch-data/')
    with open(path, 'rb') as f:
        img = Image.open(f)
        return img.convert(img.mode)
        return img.convert('RGB')

# Faster than pil_loader, if accimage is available
def accimage_loader(path):
    return  accimage.Image(path)
#     except IOError:
#         # Potentially a decoding problem, fall back to PIL.Image
#         return pil_loader(path)


def segment_panoptic(x):
    print(task, flush=True)


class LocalContrastNormalization(object):
    """
    Conduct a local contrast normalization algorithm by
    Pinto, N., Cox, D. D., and DiCarlo, J. J. (2008). Why is real-world visual object recognition hard?
     PLoS Comput Biol , 4 . 456 (they called this "Local input divisive normalization")
    the kernel size is controllable by argument kernel_size.
    """
    def __init__(self, kernel_size=2, mode='constant', cval=0.0):
        """
        :param kernel_size: int, kernel(window) size.
        :param mode: {'reflect', 'constant', 'nearest', 'mirror', 'warp'}, optional
                        determines how the array borders are handled. The meanings are listed in
                        https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.ndimage.filters.convolve.html
                        default is 'constant' as 0, different from the scipy default.
        """
        self.kernel_size = kernel_size
        self.mode = mode
        self.cval = cval

    def __call__(self, tensor):
        """
        :param tensor: Tensor image os size (C, H, W) to be normalized.
        :return:
            Tensor: Normalized Tensor image, in size (C, H, W).
        """
        C, H, W = tensor.size()
        kernel = np.ones((self.kernel_size, self.kernel_size))

        arr = np.array(tensor)
        local_sum_arr = np.array([convolve(arr[c], kernel, mode=self.mode, cval=self.cval)
                                  for c in range(C)]) # An array that has shape(C, H, W)
                                                      # Each element [c, h, w] is the summation of the values
                                                      # in the window that has arr[c,h,w] at the center.
        local_avg_arr = local_sum_arr / (self.kernel_size**2) # The tensor of local averages.

        arr_square = np.square(arr)
        local_sum_arr_square = np.array([convolve(arr_square[c], kernel, mode=self.mode, cval=self.cval)
                                  for c in range(C)]) # An array that has shape(C, H, W)
                                                      # Each element [c, h, w] is the summation of the values
                                                      # in the window that has arr_square[c,h,w] at the center.
        local_norm_arr = np.sqrt(local_sum_arr_square) # The tensor of local Euclidean norms.


        local_avg_divided_by_norm = local_avg_arr / (1e-8+local_norm_arr)

        result_arr = np.minimum(local_avg_arr, local_avg_divided_by_norm)
        return torch.Tensor(result_arr)



    def __repr__(self):
        return self.__class__.__name__ + '(kernel_size={0}, threshold={1})'.format(self.kernel_size, self.threshold)

